console.log("Better tactics");

import { Chess } from './deps/chess.js';
import { Chessground } from './deps/chessground.min.js';

const COMPUTER_MOVE_DELAY = 200;

export class Puzzle {
    constructor(container, puzzle_id, fen, moves, rating, on_success, on_move, right_move, wrong_move) {
        this._puzzle_id = puzzle_id;
        this._fen = fen;
        this._moves = moves.split(" ");
        this._rating = rating;

        this._on_success = on_success;
        this._on_move = on_move;
        this._right_move = right_move;
        this._wrong_move = wrong_move;

        this._premove = null;

        this._game = new Chess(fen);
        this._board = Chessground(container);

        // Figure out who should move first (the move that sets up the tactic), make the move,
        // and let the player make the next move.
        let first_move_origin_square = this._moves[0].slice(0, 2);

        this._computer_color = this._game.get(first_move_origin_square).color;
        this._player_color = this._computer_color == 'w' ? 'b' : 'w';

        // Start puzzle.
        this.reset();
    }

    reset() {
        this._game = new Chess(this._fen);

        this._remaining_moves = this._moves.slice();

        this._board.set({
            movable: {
                showDests: true,
                color: this._player_color == 'w' ? 'white' : 'black',
            },
            premovable: {
                events: {
                    set: this._set_premove.bind(this),
                    unset: this._unset_premove.bind(this)
                }
            },
            events: {
                move: this._move.bind(this)
            },
            orientation: this._player_color == 'w' ? 'white' : 'black',
            turnColor: this._computer_color == 'w' ? 'white' : 'black',
            fen: this._fen
        });

        setTimeout(this._make_next_move.bind(this), COMPUTER_MOVE_DELAY);
    }

    color_to_move() {
        return this._game.turn();
    }

    computer_to_move() {
        return this.color_to_move() == this._computer_color;
    }

    fen() {
        return this._game.fen();
    }

    _make_next_move() {
        if (this._game.turn() != this._computer_color) {
            console.warn("_make_next_move(): called when it's the player's turn");
            return;
        }

        // Get the next computer move.
        console.log("Making next move");
        let next_move = this._remaining_moves.shift();
        if (next_move) {
            // Get the source and destination of the move.
            let source = next_move.slice(0, 2);
            let dest = next_move.slice(2);

            // Store the last move for highlighting purposes.
            this._last_move = [source, dest];

            // Update the game and board.
            this._game.move(next_move);
            this._board.set({
                fen: this._game.fen(),
                lastMove: this._last_move,
                turnColor: this._player_color == 'w' ? 'white' : 'black'
            });
            this._board.move(next_move);

            // Call the on_move callback.
            this._on_move();

            // Add a timer to apply the next premove, if there is one.
            setTimeout(() => this._board.playPremove(), COMPUTER_MOVE_DELAY);
        }
    }

    _move(orig, dest) {
        let move = orig + dest;

        // If the puzzle is over, reject any moves.
        if (this._remaining_moves.length == 0) {
            this._board.set({
                fen: this._game.fen(),
                lastMove: this._last_move,
                turnColor: this._player_color == 'w' ? 'white' : 'black'
            });
        }

        // Make the move in chess.js, and if it was an illegal move, reject it.
        try {
            let move_result = this._game.move(move);

            // Since chessground doesn't support promotion by default, just assume auto-queen and add "q"
            // onto the end of the move so it matches the puzzle.
            // TODO: hopefully support promotion properly in the UI before coming across cases that need it.
            if (move_result.promotion) {
                move = move + "q";
                console.warn(`Encountered promotion, assuming auto-queen (${move})`);
            }
        }
        catch (_) {
            this._board.set({
                fen: this._game.fen(),
                lastMove: this._last_move,
                turnColor: this._player_color == 'w' ? 'white' : 'black'
            });
            return;
        }

        // Call on move callback now that we've validated it.
        this._on_move();

        // Check if it was the right or wrong move. A special case we need to consider is that for
        // checkmate-in-ones there might be multiple legal moves. This is the only special case
        // listed at https://database.lichess.org/#puzzles (all other moves are "only moves").
        // I originally ignored this until I found the example puzzle ULF41, in which the final
        // move is listed as Qe1# instead of the (imo) more natural Qf1#, although they are both
        // legal moves *and* checkmate.
        if (this._game.isCheckmate()) {
            this._remaining_moves.length = 0;
            this._on_success();
        }
        else if (this._remaining_moves[0] == move) {
            this._remaining_moves.shift();

            // If there are no moves left, the puzzle is complete.
            if (this._remaining_moves.length == 0) {
                this._on_success();
            }
            else {
                this._right_move();
                setTimeout(this._make_next_move.bind(this), COMPUTER_MOVE_DELAY);
            }
        }
        else {
            this._remaining_moves.length = 0;
            this._board.set({
                movable: {
                    color: 'none'
                }
            });
            this._wrong_move();
        }
    }

    _set_premove(orig, dest) {
        this._premove = [orig, dest];
    }

    _unset_premove(orig, dest) {
        this._premove = null;
    }
}
